#!/usr/bin/env python3
# SPDX-License-Identifier: ISC

# Copyright (c) 2024 Network Education Foundation, Inc. ("NetDEF")
#                    Rafael Zalamena

import os
import sys
import pytest
from functools import partial

from lib import topotest

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import Topogen, get_topogen
from lib.topolog import logger

from lib.pim import McastTesterHelper

"""
pim_embedded_rp.py: Test PIM embedded RP functionality.
"""

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# Required to instantiate the topology builder class.
pytestmark = [pytest.mark.pim6d]


def build_topo(tgen):
    tgen.add_router("r1")
    tgen.add_router("r2")
    tgen.add_router("r3")

    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])

    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r3"])

    switch = tgen.add_switch("s3")
    switch.add_link(tgen.gears["r2"])
    switch.add_link(tgen.gears["r3"])

    tgen.add_host("h1", "2001:DB8:100::100", "via 2001:DB8:100::1")
    tgen.add_host("h2", "2001:DB8:200::100", "via 2001:DB8:200::1")

    switch = tgen.add_switch("s10")
    switch.add_link(tgen.gears["r2"])
    switch.add_link(tgen.gears["h1"])

    switch = tgen.add_switch("s20")
    switch.add_link(tgen.gears["r3"])
    switch.add_link(tgen.gears["h2"])


app_helper = McastTesterHelper()


def setup_module(mod):
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()
    for _, router in router_list.items():
        router.load_frr_config(os.path.join(CWD, f"{router.name}/frr.conf"))

    tgen.start_router()
    app_helper.init(tgen)


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_ospfv3_convergence():
    "Wait for OSPFv3 protocol convergence"
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("waiting for protocols to converge")

    def expect_loopback_route(router, iptype, route, proto):
        "Wait until route is present on RIB for protocol."
        logger.info(f"waiting route {route} in {router}")
        test_func = partial(
            topotest.router_json_cmp,
            tgen.gears[router],
            f"show {iptype} route json",
            {route: [{"protocol": proto}]},
        )
        _, result = topotest.run_and_expect(test_func, None, count=10, wait=8)
        assert result is None, f'"{router}" convergence failure'

    # Wait for R1
    expect_loopback_route("r1", "ipv6", "2001:db8:ffff::2/128", "ospf6")
    expect_loopback_route("r1", "ipv6", "2001:db8:ffff::3/128", "ospf6")

    # Wait for R2
    expect_loopback_route("r2", "ipv6", "2001:db8:ffff::1/128", "ospf6")
    expect_loopback_route("r2", "ipv6", "2001:db8:ffff::3/128", "ospf6")

    # Wait for R3
    expect_loopback_route("r3", "ipv6", "2001:db8:ffff::1/128", "ospf6")
    expect_loopback_route("r3", "ipv6", "2001:db8:ffff::2/128", "ospf6")


def expect_pim_rp(router, rp, group, interface=None, missing=False):
    "Wait until RP is present."
    tgen = get_topogen()
    maximum_wait = 15
    log_message = f"waiting RP {rp} for {group} in {router}"
    if missing:
        log_message += \
            f" to be missing ({maximum_wait} seconds maximum)"

    logger.info(log_message)

    expected = {rp: [{"group": f"{group}/128"}]}
    if interface is not None:
        expected[rp][0]["outboundInterface"] = interface

    test_func = partial(
        topotest.router_json_cmp,
        tgen.gears[router],
        f"show ipv6 pim rp-info json",
        expected
    )
    _, result = topotest.run_and_expect(
        test_func, None, count=maximum_wait, wait=1)
    if missing:
        assert result is not None, f'"{router}" convergence failure'
    else:
        assert result is None, f'"{router}" convergence failure'


def test_embedded_rp_mld_join():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    success_group = "ff75:130:2001:db8:ffff::100"
    filtered_group = "ff75:330:2001:db8:ffff::200"
    app_helper.run("h1", [success_group, "h1-eth0"])
    app_helper.run("h1", [filtered_group, "h1-eth0"])

    # Expect first valid join request
    expect_pim_rp("r2", "2001:db8:ffff::1", success_group, interface="r2-eth0")

    # Expect filtered join request
    expect_pim_rp("r2", "2001:db8:ffff::2", filtered_group, missing=True)

    # Send over the limit join request
    groups = [
        "ff75:130:2001:db8:ffff::300",
        "ff75:130:2001:db8:ffff::301",
        "ff75:130:2001:db8:ffff::302",
    ]
    for group in groups:
        app_helper.run("h1", [group, "h1-eth0"])
        topotest.sleep(2, "Waiting MLD join to be sent")

    expect_pim_rp("r2", "2001:db8:ffff::1", groups[0], interface="r2-eth0")
    expect_pim_rp("r2", "2001:db8:ffff::1", groups[1], interface="r2-eth0")
    # Over the limit entry
    expect_pim_rp("r2", "2001:db8:ffff::1", groups[2], missing=True)

    app_helper.stop_all_hosts()

    # Clean up the embedded RPs so we don't cross the limit next phase
    tgen.gears["r2"].vtysh_cmd("clear ipv6 mroute")


def test_embedded_rp_pim_join():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    #
    # Test sending PIM join with embedded RP information to router R2
    #
    group = "ff75:230:2001:db8:ffff::400"
    app_helper.run("h2", [group, "h2-eth0"])
    expect_pim_rp("r3", "2001:db8:ffff::2", group, interface="r3-eth1")
    expect_pim_rp("r2", "2001:db8:ffff::2", group, interface="lo")

    app_helper.stop_all_hosts()


def test_embedded_rp_spt_switch():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Clean up the embedded RPs so we don't cross the limit next phase
    tgen.gears["r1"].vtysh_cmd("clear ipv6 mroute")
    tgen.gears["r2"].vtysh_cmd("clear ipv6 mroute")
    tgen.gears["r3"].vtysh_cmd("clear ipv6 mroute")

    group = "ff75:130:2001:db8:ffff::500"
    rp = "2001:db8:ffff::1"
    source = "2001:db8:100::100"

    # Join from r3 (host h2)
    app_helper.run("h2", [group, "h2-eth0"])
    # Wait for embedded RP to show up
    expect_pim_rp("r3", rp, group, interface="r3-eth0")

    # Send stream from r2 (host h1)
    app_helper.run("h1", ["--send=0.7", group, "h1-eth0"])

    # Check if R1 has the correct multicast route
    logger.info("Waiting r1 multicast route installation")
    test_func = partial(
        topotest.router_json_cmp,
        tgen.gears["r1"],
        f"show ipv6 pim state json",
        {group: {"*": {}, source: {}}}
    )
    _, result = topotest.run_and_expect(test_func, None, count=10, wait=8)
    assert result is None, '"r1" convergence failure'

    # Check if R2 has the correct multicast route
    logger.info("Waiting r2 multicast route installation")
    test_func = partial(
        topotest.router_json_cmp,
        tgen.gears["r2"],
        f"show ipv6 pim state json",
        {group: {source: {"r2-eth2": {"r2-eth1": {}}}}}
    )
    _, result = topotest.run_and_expect(test_func, None, count=10, wait=8)
    assert result is None, '"r2" convergence failure'

    # Check if R3 has the correct multicast route
    logger.info("Waiting r3 multicast route installation")
    test_func = partial(
        topotest.router_json_cmp,
        tgen.gears["r3"],
        f"show ipv6 pim state json",
        {group: {source: {"r3-eth1": {"r3-eth2": {}}}}}
    )
    _, result = topotest.run_and_expect(test_func, None, count=10, wait=8)
    assert result is None, '"r3" convergence failure'


def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2022 by VMware, Inc. ("VMware")
# Used Copyright (c) 2018 by Network Device Education Foundation,
# Inc. ("NetDEF") in this file.
#

import sys
import time
import pytest
import inspect
import os


"""Following tests are covered to test bgp admin distance functionality.
TC_1:
    Verify bgp admin distance functionality when static route is
    configured same as ebgp learnt route

TC_2:
    Verify ebgp admin distance functionality with ECMP.

TC_3:
    Verify ibgp admin distance functionality when static route is
    configured same as bgp learnt route.
TC_4:
    Verify ibgp admin distance functionality with ECMP.

TC_7: Chaos - Verify bgp admin distance functionality with chaos.
"""

#################################
# TOPOLOGY
#################################
"""

                    +-------+
         +--------- |  R2   |
         |          +-------+
         |iBGP           |
     +-------+           |
     |  R1   |           |iBGP
     +-------+           |
         |               |
         |    iBGP   +-------+   eBGP   +-------+
         +---------- |  R3   |----------|  R4   |
                     +-------+          +-------+
                        |
                        |eBGP
                        |
                    +-------+
                    |  R5   |
                    +-------+


"""

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))
sys.path.append(os.path.join(CWD, "../lib/"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import Topogen, get_topogen

# Required to instantiate the topology builder class.
from lib.common_config import (
    start_topology,
    write_test_header,
    step,
    write_test_footer,
    create_static_routes,
    verify_rib,
    create_route_maps,
    create_prefix_lists,
    check_address_types,
    reset_config_on_routers,
    check_router_status,
    stop_router,
    kill_router_daemons,
    start_router_daemons,
    start_router,
    get_frr_ipv6_linklocal,
    verify_fib_routes,
)
from lib.topolog import logger
from lib.bgp import (
    verify_bgp_convergence,
    create_router_bgp,
    verify_best_path_as_per_admin_distance,
    clear_bgp,
)

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import Topogen, get_topogen
from lib.topojson import build_config_from_json
from lib.topolog import logger

# Global variables
topo = None
bgp_convergence = False
pytestmark = [pytest.mark.bgpd, pytest.mark.staticd]

NETWORK = {
    "ipv4": [
        "192.168.20.1/32",
        "192.168.20.2/32",
        "192.168.21.1/32",
        "192.168.21.2/32",
        "192.168.22.1/32",
        "192.168.22.2/32",
    ],
    "ipv6": [
        "fc07:50::1/128",
        "fc07:50::2/128",
        "fc07:150::1/128",
        "fc07:150::2/128",
        "fc07:1::1/128",
        "fc07:1::2/128",
    ],
}

ADDR_TYPES = check_address_types()


def setup_module(mod):
    """
    Sets up the pytest environment

    * `mod`: module name
    """

    global topo

    testsuite_run_time = time.asctime(time.localtime(time.time()))
    logger.info("Testsuite start time: {}".format(testsuite_run_time))
    logger.info("=" * 40)

    logger.info("Running setup_module to create topology")

    # This function initiates the topology build with Topogen...
    json_file = "{}/bgp_admin_dist.json".format(CWD)
    tgen = Topogen(json_file, mod.__name__)
    global topo
    topo = tgen.json_topo

    # Starting topology, create tmp files which are loaded to routers
    #  to start deamons and then start routers
    start_topology(tgen)

    # Creating configuration from JSON
    build_config_from_json(tgen, topo)

    # Checking BGP convergence
    global bgp_convergence
    global ADDR_TYPES

    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Api call verify whether BGP is converged
    bgp_convergence = verify_bgp_convergence(tgen, topo)
    assert bgp_convergence is True, "setup_module :Failed \n Error:" " {}".format(
        bgp_convergence
    )
    logger.info("Running setup_module() done")


def teardown_module(mod):
    """teardown_module.

    Teardown the pytest environment.
    * `mod`: module name
    """
    logger.info("Running teardown_module to delete topology")
    tgen = get_topogen()

    # Stop toplogy and Remove tmp files
    tgen.stop_topology()

    logger.info(
        "Testsuite end time: {}".format(time.asctime(time.localtime(time.time())))
    )
    logger.info("=" * 40)


#####################################################
# Tests starting
#####################################################
def test_bgp_admin_distance_ebgp_ecmp_p0():
    """
    TC: 2
    Verify ebgp admin distance functionality with ECMP.
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip("skipping test case because of BGP Convergence failure at setup")

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    if tgen.routers_have_failure():
        check_router_status(tgen)

    step("Configure base config as per the topology")
    reset_config_on_routers(tgen)

    step("Configure static route  in R4 and R5, redistribute in bgp")

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r4": {
                "static_routes": [{"network": NETWORK[addr_type], "next_hop": "Null0"}]
            }
        }

        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r5": {
                "static_routes": [{"network": NETWORK[addr_type], "next_hop": "Null0"}]
            }
        }

        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Verify that route is learnt in DUT via ebgp")

    # Verifying RIB routes
    protocol = "bgp"
    input_dict = topo["routers"]
    dut = "r3"
    nhop = {"ipv4": [], "ipv6": []}
    nhop["ipv4"].append(topo["routers"]["r4"]["links"]["r3"]["ipv4"].split("/")[0])
    nhop["ipv4"].append(topo["routers"]["r5"]["links"]["r3"]["ipv4"].split("/")[0])
    nhop["ipv6"].append(get_frr_ipv6_linklocal(tgen, "r4", "r3-r4-eth1"))
    nhop["ipv6"].append(get_frr_ipv6_linklocal(tgen, "r5", "r1-r3-eth1"))

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {"network": NETWORK[addr_type][0], "next_hop": "Null0"}
                ]
            }
        }
        result4 = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    step("Configure the static route  in R3 (Dut).")

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {"network": NETWORK[addr_type][0], "next_hop": "Null0"}
                ]
            }
        }

        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Verify that static route is selected as best route in zebra.")

    # Verifying RIB routes
    protocol = "static"
    dut = "r3"

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {"network": NETWORK[addr_type][0], "next_hop": "Null0"}
                ]
            }
        }

        result4 = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    step(" Configure the admin distance of 254 to static route  in R3.")

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type][0],
                        "next_hop": "Null0",
                        "admin_distance": 254,
                    }
                ]
            }
        }

        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Verify that bgp routes are selected as best routes in zebra.")
    protocol = "bgp"
    dut = "r3"

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {"network": NETWORK[addr_type][0], "next_hop": "Null0"}
                ]
            }
        }
        result4 = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    input_dict_1 = {
        "r3": {
            "bgp": {
                "local_as": 100,
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "distance": {"ebgp": 254, "ibgp": 254, "local": 254}
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "distance": {"ebgp": 254, "ibgp": 254, "local": 254}
                        }
                    },
                },
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify that bgp routes are selected as best routes in zebra.")
    # Verifying RIB routes
    protocol = "bgp"
    dut = "r3"

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {"network": NETWORK[addr_type][0], "next_hop": "Null0"}
                ]
            }
        }
        result4 = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    step("Configure bgp admin distance 10 with CLI in dut.")
    input_dict_1 = {
        "r3": {
            "bgp": {
                "local_as": 100,
                "address_family": {
                    "ipv4": {
                        "unicast": {"distance": {"ebgp": 10, "ibgp": 254, "local": 254}}
                    },
                    "ipv6": {
                        "unicast": {"distance": {"ebgp": 10, "ibgp": 254, "local": 254}}
                    },
                },
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify ebgp routes have admin distance of 10 in dut.")

    protocol = "bgp"
    input_dict = topo["routers"]
    dut = "r3"

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {"network": NETWORK[addr_type][0], "next_hop": "Null0"}
                ]
            }
        }
        result4 = verify_rib(
            tgen, addr_type, dut, input_dict, protocol=protocol, admin_distance=10
        )
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    step(
        "Configure route map with weight as 200 and apply to one of the "
        "neighbor (R4 neighbor)."
    )

    # Create Prefix list
    input_dict_2 = {
        "r3": {
            "prefix_lists": {
                "ipv4": {
                    "pf_ls_1": [
                        {
                            "seqid": 10,
                            "network": NETWORK["ipv4"][0],
                            "le": "32",
                            "action": "permit",
                        }
                    ]
                },
                "ipv6": {
                    "pf_ls_1_ipv6": [
                        {
                            "seqid": 100,
                            "network": NETWORK["ipv6"][0],
                            "le": "128",
                            "action": "permit",
                        }
                    ]
                },
            }
        }
    }
    result = create_prefix_lists(tgen, input_dict_2)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    # Create route map
    input_dict_3 = {
        "r3": {
            "route_maps": {
                "RMAP_WEIGHT": [
                    {
                        "action": "permit",
                        "match": {"ipv4": {"prefix_lists": "pf_ls_1"}},
                        "set": {"weight": 200},
                    },
                    {
                        "action": "permit",
                        "match": {"ipv6": {"prefix_lists": "pf_ls_1_ipv6"}},
                        "set": {"weight": 200},
                    },
                ]
            }
        }
    }
    result = create_route_maps(tgen, input_dict_3)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    # Configure neighbor for route map
    input_dict_4 = {
        "r3": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r4": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [
                                                {
                                                    "name": "RMAP_WEIGHT",
                                                    "direction": "in",
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r4": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [
                                                {
                                                    "name": "RMAP_WEIGHT",
                                                    "direction": "in",
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        }
    }
    result = create_router_bgp(tgen, topo, input_dict_4)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify that bgp route is selected as best on by zebra in r3.")

    protocol = "bgp"
    dut = "r3"

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {"network": NETWORK[addr_type][0], "next_hop": "Null0"}
                ]
            }
        }
        result4 = verify_rib(
            tgen, addr_type, dut, input_dict, protocol=protocol, admin_distance=10
        )
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    step("Static route should not be selected as best route.")
    protocol = "static"
    dut = "r3"

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {"network": NETWORK[addr_type][0], "next_hop": "Null0"}
                ]
            }
        }
        result4 = verify_fib_routes(
            tgen, addr_type, dut, input_dict, protocol=protocol, expected=False
        )
        assert (
            result4 is not True
        ), "Testcase {} : Failed. Wrong route is selected as best route.\n Error: {}".format(
            tc_name, result4
        )

    step("Reconfigure the static route without admin distance")

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type][0],
                        "next_hop": "Null0",
                        "admin_distance": 254,
                        "delete": True,
                    }
                ]
            }
        }

        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {"network": NETWORK[addr_type][0], "next_hop": "Null0"}
                ]
            }
        }

        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Verify that static route is installed as best route.")
    protocol = "static"
    dut = "r3"

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {"network": NETWORK[addr_type][0], "next_hop": "Null0"}
                ]
            }
        }
        result4 = verify_rib(
            tgen, addr_type, dut, input_dict, protocol=protocol, fib=True
        )
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    step("Unconfigure the static route in R3.")

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type][0],
                        "next_hop": "Null0",
                        "delete": True,
                    }
                ]
            }
        }

        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Verify that bgp route is selected as best on by zebra in r3.")

    protocol = "bgp"
    dut = "r3"

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {"network": NETWORK[addr_type][0], "next_hop": "Null0"}
                ]
            }
        }
        result4 = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    step("Un configure the route map on R3.")

    # Configure neighbor for route map
    input_dict_4 = {
        "r3": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r4": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [
                                                {
                                                    "name": "RMAP_WEIGHT",
                                                    "direction": "in",
                                                    "delete": True,
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r4": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [
                                                {
                                                    "name": "RMAP_WEIGHT",
                                                    "direction": "in",
                                                    "delete": True,
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        }
    }
    result = create_router_bgp(tgen, topo, input_dict_4)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("verify bgp routes installed in zebra.")

    # Verifying RIB routes
    protocol = "bgp"
    input_dict = topo["routers"]
    dut = "r3"
    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {"network": NETWORK[addr_type][0], "next_hop": "Null0"}
                ]
            }
        }
        result4 = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    write_test_footer(tc_name)


def test_bgp_admin_distance_ibgp_p0():
    """
    TC: 3
    Verify bgp admin distance functionality when static route is
    configured same as ibgp learnt route
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip("skipping test case because of BGP Convergence failure at setup")

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    if tgen.routers_have_failure():
        check_router_status(tgen)

    step("Configure base config as per the topology")
    reset_config_on_routers(tgen)

    step("Configure bgp admin distance 200 with CLI in dut.")

    input_dict_1 = {
        "r3": {
            "bgp": {
                "local_as": 100,
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "distance": {"ebgp": 200, "ibgp": 200, "local": 200}
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "distance": {"ebgp": 200, "ibgp": 200, "local": 200}
                        }
                    },
                },
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify bgp routes have admin distance of 200 in dut.")
    # Verifying best path
    dut = "r3"
    attribute = "admin_distance"

    input_dict = {
        "ipv4": {
            "r3": {
                "static_routes": [
                    {
                        "network": "192.168.22.1/32",
                        "admin_distance": 200,
                    },
                    {
                        "network": "192.168.22.2/32",
                        "admin_distance": 200,
                    },
                ]
            }
        },
        "ipv6": {
            "r3": {
                "static_routes": [
                    {
                        "network": "fc07:1::1/128",
                        "admin_distance": 200,
                    },
                    {
                        "network": "fc07:1::2/128",
                        "admin_distance": 200,
                    },
                ]
            }
        },
    }

    for addr_type in ADDR_TYPES:
        result = verify_best_path_as_per_admin_distance(
            tgen, addr_type, dut, input_dict[addr_type], attribute
        )
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Modify the admin distance value to 150.")

    input_dict_1 = {
        "r3": {
            "bgp": {
                "local_as": 100,
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "distance": {"ebgp": 150, "ibgp": 150, "local": 150}
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "distance": {"ebgp": 150, "ibgp": 150, "local": 150}
                        }
                    },
                },
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify bgp routes have admin distance of 150 in dut.")
    # Verifying best path
    dut = "r3"
    attribute = "admin_distance"

    input_dict = {
        "ipv4": {
            "r3": {
                "static_routes": [
                    {
                        "network": "192.168.22.1/32",
                        "admin_distance": 150,
                    },
                    {
                        "network": "192.168.22.2/32",
                        "admin_distance": 150,
                    },
                ]
            }
        },
        "ipv6": {
            "r3": {
                "static_routes": [
                    {
                        "network": "fc07:1::1/128",
                        "admin_distance": 150,
                    },
                    {
                        "network": "fc07:1::2/128",
                        "admin_distance": 150,
                    },
                ]
            }
        },
    }

    for addr_type in ADDR_TYPES:
        result = verify_best_path_as_per_admin_distance(
            tgen, addr_type, dut, input_dict[addr_type], attribute
        )
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Un configure the admin distance value on DUT")

    input_dict_1 = {
        "r3": {
            "bgp": {
                "local_as": 100,
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "distance": {
                                "ebgp": 150,
                                "ibgp": 150,
                                "local": 150,
                                "delete": True,
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "distance": {
                                "ebgp": 150,
                                "ibgp": 150,
                                "local": 150,
                                "delete": True,
                            }
                        }
                    },
                },
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify bgp routes have default admin distance in dut.")
    # Verifying best path
    dut = "r3"
    attribute = "admin_distance"

    input_dict = {
        "ipv4": {
            "r3": {
                "static_routes": [
                    {
                        "network": "192.168.22.1/32",
                        "admin_distance": 20,
                    },
                    {
                        "network": "192.168.22.2/32",
                        "admin_distance": 20,
                    },
                ]
            }
        },
        "ipv6": {
            "r3": {
                "static_routes": [
                    {
                        "network": "fc07:1::1/128",
                        "admin_distance": 20,
                    },
                    {
                        "network": "fc07:1::2/128",
                        "admin_distance": 20,
                    },
                ]
            }
        },
    }

    for addr_type in ADDR_TYPES:
        result = verify_best_path_as_per_admin_distance(
            tgen, addr_type, dut, input_dict[addr_type], attribute
        )
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step(
        "Learn the same route via ebgp and ibgp peer. Configure admin "
        "distance of 200 in DUT for both ebgp and ibgp peer. "
    )

    step("Verify that ebgp route is preferred over ibgp.")

    # Verifying RIB routes
    protocol = "bgp"
    input_dict = topo["routers"]

    for addr_type in ADDR_TYPES:
        result4 = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    step("Configure static route  Without any admin distance")

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [{"network": NETWORK[addr_type], "next_hop": "Null0"}]
            }
        }

        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Verify that zebra selects static route.")
    protocol = "static"

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [{"network": NETWORK[addr_type], "next_hop": "Null0"}]
            }
        }

        result4 = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    step("Configure static route  with admin distance of 253")
    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": "Null0",
                        "admin_distance": 253,
                    }
                ]
            }
        }

        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Verify that zebra selects bgp route.")
    protocol = "bgp"

    for addr_type in ADDR_TYPES:
        result4 = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    step("Configure admin distance of 254 in bgp for route.")

    input_dict_1 = {
        "r3": {
            "bgp": {
                "local_as": 100,
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "distance": {"ebgp": 254, "ibgp": 254, "local": 254}
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "distance": {"ebgp": 254, "ibgp": 254, "local": 254}
                        }
                    },
                },
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify that zebra selects static route.")
    protocol = "static"

    for addr_type in ADDR_TYPES:
        result4 = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    step("Delete the static route.")
    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": "Null0",
                        "admin_distance": 253,
                        "delete": True,
                    }
                ]
            }
        }

        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Verify that zebra selects bgp route.")
    protocol = "bgp"

    for addr_type in ADDR_TYPES:
        result4 = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    write_test_footer(tc_name)


def test_bgp_admin_distance_chaos_p2():
    """
    TC: 7
    Chaos - Verify bgp admin distance functionality with chaos.
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip("skipping test case because of BGP Convergence failure at setup")

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    if tgen.routers_have_failure():
        check_router_status(tgen)

    step("Configure base config as per the topology")
    reset_config_on_routers(tgen)

    step("Configure bgp admin distance 200 with CLI in dut.")

    input_dict_1 = {
        "r3": {
            "bgp": {
                "local_as": 100,
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "distance": {"ebgp": 200, "ibgp": 200, "local": 200}
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "distance": {"ebgp": 200, "ibgp": 200, "local": 200}
                        }
                    },
                },
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify bgp routes have admin distance of 200 in dut.")
    # Verifying best path
    dut = "r3"
    attribute = "admin_distance"

    input_dict = {
        "ipv4": {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK["ipv4"][0],
                        "admin_distance": 200,
                    },
                    {
                        "network": NETWORK["ipv4"][1],
                        "admin_distance": 200,
                    },
                ]
            }
        },
        "ipv6": {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK["ipv6"][0],
                        "admin_distance": 200,
                    },
                    {
                        "network": NETWORK["ipv6"][1],
                        "admin_distance": 200,
                    },
                ]
            }
        },
    }

    for addr_type in ADDR_TYPES:
        result = verify_best_path_as_per_admin_distance(
            tgen, addr_type, dut, input_dict[addr_type], attribute
        )
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Restart frr on R3")
    stop_router(tgen, "r3")
    start_router(tgen, "r3")

    bgp_convergence = verify_bgp_convergence(tgen, topo)
    assert bgp_convergence is True, "Testcase {} : Failed \n Error: {}".format(
        tc_name, bgp_convergence
    )

    step("Verify ebgp and ibgp routes have admin distance of 200 in dut.")
    for addr_type in ADDR_TYPES:
        result = verify_best_path_as_per_admin_distance(
            tgen, addr_type, dut, input_dict[addr_type], attribute
        )
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Restart bgpd process on R3")
    kill_router_daemons(tgen, "r3", ["bgpd"])
    start_router_daemons(tgen, "r3", ["bgpd"])

    bgp_convergence = verify_bgp_convergence(tgen, topo)
    assert bgp_convergence is True, "Testcase {} : Failed \n Error: {}".format(
        tc_name, bgp_convergence
    )

    step("Verify ebgp and ibgp routes have admin distance of 200 in dut.")
    for addr_type in ADDR_TYPES:
        result = verify_best_path_as_per_admin_distance(
            tgen, addr_type, dut, input_dict[addr_type], attribute
        )
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Clear BGP")
    for rtr in topo["routers"]:
        clear_bgp(tgen, "ipv4", rtr)
        clear_bgp(tgen, "ipv6", rtr)

    bgp_convergence = verify_bgp_convergence(tgen, topo)
    assert bgp_convergence is True, "Testcase {} : Failed \n Error: {}".format(
        tc_name, bgp_convergence
    )

    step("Verify that zebra selects bgp route.")
    protocol = "bgp"

    for addr_type in ADDR_TYPES:
        result4 = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    write_test_footer(tc_name)


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



# -*- coding: utf-8 eval: (blacken-mode 1) -*-
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2022 by VMware, Inc. ("VMware")
# Used Copyright (c) 2018 by Network Device Education Foundation,
# Inc. ("NetDEF") in this file.
#

"""
Following tests are covered to test Multicast basic functionality:

Topology:

                 _______r2_____
                |             |
      iperf     |             |     iperf
        r0-----r1-------------r3-----r5
                |             |
                |_____________|
                        r4

Test steps
- Create topology (setup module)
- Bring up topology

1. Configure multiple  groups (10 grps) with same RP address
2. Verify IIF and OIL in updated in mroute when upstream interface
   configure as RP
3. Verify RP info and (*,G) mroute after deleting the RP and shut /
    no shut the RPF interface.
"""

import os
import sys
import time

import pytest
from lib.common_config import (
    create_debug_log_config,
    reset_config_on_routers,
    shutdown_bringup_interface,
    start_topology,
    step,
    write_test_footer,
    write_test_header,
)
from lib.pim import (
    McastTesterHelper,
    create_pim_config,
    verify_join_state_and_timer,
    verify_mld_groups,
    verify_mroutes,
    verify_pim6_neighbors,
    verify_pim_rp_info,
    verify_upstream_iif,
)
from lib.topogen import Topogen, get_topogen
from lib.topojson import build_config_from_json, build_topo_from_json
from lib.topolog import logger

# Global variables
GROUP_RANGE_1 = "ff08::/64"
GROUP_ADDRESS_1 = "ff08::1"
GROUP_RANGE_3 = "ffaa::/64"
GROUP_ADDRESS_3 = "ffaa::1"
GROUP_RANGE_LIST_1 = [
    "ffaa::1/128",
    "ffaa::2/128",
    "ffaa::3/128",
    "ffaa::4/128",
    "ffaa::5/128",
]
GROUP_RANGE_LIST_2 = [
    "ffaa::6/128",
    "ffaa::7/128",
    "ffaa::8/128",
    "ffaa::9/128",
    "ffaa::10/128",
]
GROUP_ADDRESS_LIST_1 = ["ffaa::1", "ffaa::2", "ffaa::3", "ffaa::4", "ffaa::5"]
GROUP_ADDRESS_LIST_2 = ["ffaa::6", "ffaa::7", "ffaa::8", "ffaa::9", "ffaa::10"]
STAR = "*"
SOURCE = "Static"
ASSERT_MSG = "Testcase {} : Failed Error: {}"

pytestmark = [pytest.mark.pim6d]


def build_topo(tgen):
    """Build function"""

    # Building topology from json file
    build_topo_from_json(tgen, TOPO)


def setup_module(mod):
    """
    Sets up the pytest environment

    * `mod`: module name
    """

    testsuite_run_time = time.asctime(time.localtime(time.time()))
    logger.info("Testsuite start time: %s", testsuite_run_time)
    logger.info("=" * 40)

    topology = """

                 _______r2_____
                |             |
      iperf     |             |     iperf
        r0-----r1-------------r3-----r5
                |             |
                |_____________|
                        r4

    """
    logger.info("Master Topology: \n %s", topology)

    logger.info("Running setup_module to create topology")

    # This function initiates the topology build with Topogen...
    json_file = "multicast_pim6_static_rp.json"
    tgen = Topogen(json_file, mod.__name__)
    global TOPO
    TOPO = tgen.json_topo

    # ... and here it calls Mininet initialization functions.

    # Starting topology, create tmp files which are loaded to routers
    #  to start daemons and then start routers
    start_topology(tgen)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    build_config_from_json(tgen, TOPO)

    # Verify PIM6 neighbors
    result = verify_pim6_neighbors(tgen, TOPO)
    assert result is True, "setup_module :Failed \n Error:" " {}".format(result)

    global app_helper
    app_helper = McastTesterHelper(tgen)

    logger.info("Running setup_module() done")


def teardown_module():
    """Teardown the pytest environment"""

    logger.info("Running teardown_module to delete topology")
    tgen = get_topogen()

    app_helper.cleanup()

    # Stop toplogy and Remove tmp files
    tgen.stop_topology()

    logger.info("Testsuite end time: %s", time.asctime(time.localtime(time.time())))
    logger.info("=" * 40)


#####################################################
#
#   Local API
#
#####################################################


def verify_state_incremented(state_before, state_after):
    """
    API to compare interface traffic state incrementing

    Parameters
    ----------
    * `state_before` : State dictionary for any particular instance
    * `state_after` : State dictionary for any particular instance
    """

    for router, state_data in state_before.items():
        for state, value in state_data.items():
            if state_before[router][state] >= state_after[router][state]:
                errormsg = (
                    "[DUT: %s]: state %s value has not"
                    " incremented, Initial value: %s, "
                    "Current value: %s [FAILED!!]"
                    % (
                        router,
                        state,
                        state_before[router][state],
                        state_after[router][state],
                    )
                )
                return errormsg

            logger.info(
                "[DUT: %s]: State %s value is "
                "incremented, Initial value: %s, Current value: %s"
                " [PASSED!!]",
                router,
                state,
                state_before[router][state],
                state_after[router][state],
            )

    return True


#####################################################
#
#   Testcases
#
#####################################################


def test_pim6_multiple_groups_same_RP_address_p2(request):
    """
    Configure multiple  groups (10 grps) with same RP address

    Topology used:
                ________r2_____
                |             |
      iperf     |             |     iperf
        r0-----r1-------------r3-----r5

    r1 : LHR
    r2 : RP
    r3 : FHR
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("Creating configuration from JSON")
    reset_config_on_routers(tgen)

    app_helper.stop_all_hosts()

    input_dict = {
        "r1": {"debug": {"log_file": "r1_debug.log", "enable": ["pim6d"]}},
        "r2": {"debug": {"log_file": "r2_debug.log", "enable": ["pim6d"]}},
        "r3": {"debug": {"log_file": "r3_debug.log", "enable": ["pim6d"]}},
        "r4": {"debug": {"log_file": "r4_debug.log", "enable": ["pim6d"]}},
    }

    result = create_debug_log_config(tgen, input_dict)

    step("Enable MLD on r1 interface")
    step("Enable the PIM6 on all the interfaces of r1, r2, r3 and r4 routers")
    step("r2: Configure r2 as RP")
    input_dict = {
        "r2": {
            "pim6": {
                "rp": [
                    {
                        "rp_addr": TOPO["routers"]["r2"]["links"]["lo"]["ipv6"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_3,
                    }
                ]
            }
        }
    }

    result = create_pim_config(tgen, TOPO, input_dict)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("r2: verify rp-info")
    dut = "r2"
    oif = "lo"
    rp_address = TOPO["routers"]["r2"]["links"]["lo"]["ipv6"].split("/")[0]
    result = verify_pim_rp_info(tgen, TOPO, dut, GROUP_RANGE_3, oif, rp_address, SOURCE)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    group_address_list = GROUP_ADDRESS_LIST_1 + GROUP_ADDRESS_LIST_2
    step("r0: Send MLD join for 10 groups")
    intf = TOPO["routers"]["r0"]["links"]["r1"]["interface"]
    result = app_helper.run_join("r0", group_address_list, "r1")
    assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    step("r1: Verify MLD groups")
    dut = "r1"
    intf_r1_r0 = TOPO["routers"]["r1"]["links"]["r0"]["interface"]
    result = verify_mld_groups(tgen, dut, intf_r1_r0, group_address_list)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r5: Send multicast traffic for group {}".format(group_address_list))
    SOURCE_ADDRESS = TOPO["routers"]["r5"]["links"]["r3"]["ipv6"].split("/")[0]
    result = app_helper.run_traffic("r5", group_address_list, "r3")
    assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    step("r1: Verify (*, G) upstream IIF interface")
    dut = "r1"
    iif = TOPO["routers"]["r1"]["links"]["r2"]["interface"]
    result = verify_upstream_iif(tgen, dut, iif, STAR, group_address_list)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (*, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, STAR, group_address_list, addr_type="ipv6"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (*, G) ip mroutes")
    oif = TOPO["routers"]["r1"]["links"]["r0"]["interface"]
    result = verify_mroutes(tgen, dut, STAR, group_address_list, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (S, G) ip mroutes")
    iif = TOPO["routers"]["r1"]["links"]["r3"]["interface"]
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, group_address_list, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (S, G) upstream IIF interface")
    result = verify_upstream_iif(tgen, dut, iif, SOURCE_ADDRESS, group_address_list)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, SOURCE_ADDRESS, group_address_list, addr_type="ipv6"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (*, G) upstream IIF interface")
    dut = "r2"
    iif = "lo"
    result = verify_upstream_iif(tgen, dut, iif, STAR, group_address_list)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (*, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, STAR, group_address_list, addr_type="ipv6"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (*, G) ip mroutes")
    oif = TOPO["routers"]["r2"]["links"]["r1"]["interface"]
    result = verify_mroutes(tgen, dut, STAR, group_address_list, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r3: Verify (S, G) upstream IIF interface")
    dut = "r3"
    iif = TOPO["routers"]["r3"]["links"]["r5"]["interface"]
    result = verify_upstream_iif(tgen, dut, iif, SOURCE_ADDRESS, group_address_list)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r3: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen,
        dut,
        iif,
        SOURCE_ADDRESS,
        group_address_list,
        addr_type="ipv6",
        expected=False,
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "r3: (S,G) upstream state is joined and join timer is running\n Error: {}".format(
            tc_name, result
        )
    )

    step("r3: Verify (S, G) ip mroutes")
    oif = TOPO["routers"]["r3"]["links"]["r1"]["interface"]
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, group_address_list, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (S, G) upstream IIF interface")
    dut = "r2"
    iif = TOPO["routers"]["r2"]["links"]["r3"]["interface"]
    result = verify_upstream_iif(
        tgen, dut, iif, SOURCE_ADDRESS, group_address_list, joinState="NotJoined"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen,
        dut,
        iif,
        SOURCE_ADDRESS,
        group_address_list,
        addr_type="ipv6",
        expected=False,
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "r2: (S,G) upstream state is joined and join timer is running\n Error: {}".format(
            tc_name, result
        )
    )

    step("r2: Verify (S, G) ip mroutes")
    oif = "none"
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, group_address_list, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Delete RP configuration")
    input_dict = {
        "r1": {
            "pim6": {
                "rp": [
                    {
                        "rp_addr": TOPO["routers"]["r2"]["links"]["lo"]["ipv6"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_3,
                        "delete": True,
                    }
                ]
            }
        }
    }

    result = create_pim_config(tgen, TOPO, input_dict)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("r1: Shut the interface r1-r2-eth1 from R1 to R2")
    dut = "r1"
    intf = TOPO["routers"]["r1"]["links"]["r2"]["interface"]
    shutdown_bringup_interface(tgen, dut, intf, False)

    step("r1: No Shut the interface r1-r2-eth1 from R1 to R2")
    shutdown_bringup_interface(tgen, dut, intf, True)

    step("r1: Configure RP")
    input_dict = {
        "r1": {
            "pim6": {
                "rp": [
                    {
                        "rp_addr": TOPO["routers"]["r2"]["links"]["lo"]["ipv6"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_3,
                    }
                ]
            }
        }
    }

    result = create_pim_config(tgen, TOPO, input_dict)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("r1: Shut the interface r1-r0-eth0 from R1 to R2")
    intf = TOPO["routers"]["r1"]["links"]["r0"]["interface"]
    shutdown_bringup_interface(tgen, dut, intf, False)

    step("r1: No Shut the interface r1-r0-eth0 from R1 to R2")
    shutdown_bringup_interface(tgen, dut, intf, True)

    step("r1: Verify (*, G) upstream IIF interface")
    dut = "r1"
    iif = TOPO["routers"]["r1"]["links"]["r2"]["interface"]
    result = verify_upstream_iif(tgen, dut, iif, STAR, group_address_list)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (*, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, STAR, group_address_list, addr_type="ipv6"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (*, G) ip mroutes")
    oif = TOPO["routers"]["r1"]["links"]["r0"]["interface"]
    result = verify_mroutes(tgen, dut, STAR, group_address_list, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (S, G) upstream IIF interface")
    iif = TOPO["routers"]["r1"]["links"]["r3"]["interface"]
    result = verify_upstream_iif(tgen, dut, iif, SOURCE_ADDRESS, group_address_list)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, SOURCE_ADDRESS, group_address_list, addr_type="ipv6"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (S, G) ip mroutes")
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, group_address_list, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (*, G) upstream IIF interface")
    dut = "r2"
    iif = "lo"
    result = verify_upstream_iif(tgen, dut, iif, STAR, group_address_list)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (*, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, STAR, group_address_list, addr_type="ipv6"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (*, G) ip mroutes")
    oif = TOPO["routers"]["r2"]["links"]["r1"]["interface"]
    result = verify_mroutes(tgen, dut, STAR, group_address_list, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r3: Verify (S, G) upstream IIF interface")
    dut = "r3"
    iif = TOPO["routers"]["r3"]["links"]["r5"]["interface"]
    result = verify_upstream_iif(tgen, dut, iif, SOURCE_ADDRESS, group_address_list)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r3: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen,
        dut,
        iif,
        SOURCE_ADDRESS,
        group_address_list,
        addr_type="ipv6",
        expected=False,
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "r3: (S,G) upstream state is joined and join timer is running\n Error: {}".format(
            tc_name, result
        )
    )

    step("r3: Verify (S, G) ip mroutes")
    oif = TOPO["routers"]["r3"]["links"]["r1"]["interface"]
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, group_address_list, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    write_test_footer(tc_name)


def test_pim6_multiple_groups_different_RP_address_p2(request):
    """
    Verify IIF and OIL in updated in mroute when upstream interface
    configure as RP

    Topology used:
                ________r2_____
                |             |
      iperf     |             |     iperf
        r0-----r1-------------r3-----r5
                |             |
                |_____________|
                        r4
    r1 : LHR
    r2 & r4 : RP
    r3 : FHR
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("Creating configuration from JSON")
    reset_config_on_routers(tgen)

    app_helper.stop_all_hosts()

    step("Enable MLD on r1 interface")
    step("Enable the PIM6 on all the interfaces of r1, r2, r3 and r4 routers")
    step("r2: Configure r2 as RP")
    step("r4: Configure r4 as RP")
    input_dict = {
        "r2": {
            "pim6": {
                "rp": [
                    {
                        "rp_addr": TOPO["routers"]["r2"]["links"]["lo"]["ipv6"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_LIST_1,
                    }
                ]
            }
        },
        "r4": {
            "pim6": {
                "rp": [
                    {
                        "rp_addr": TOPO["routers"]["r4"]["links"]["lo"]["ipv6"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_LIST_2,
                    }
                ]
            }
        },
    }

    result = create_pim_config(tgen, TOPO, input_dict)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("r2: Verify RP info")
    dut = "r2"
    oif = "lo"
    rp_address = TOPO["routers"]["r2"]["links"]["lo"]["ipv6"].split("/")[0]
    result = verify_pim_rp_info(
        tgen, TOPO, dut, GROUP_RANGE_LIST_1, oif, rp_address, SOURCE
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r4: Verify RP info")
    dut = "r4"
    rp_address = TOPO["routers"]["r4"]["links"]["lo"]["ipv6"].split("/")[0]
    result = verify_pim_rp_info(
        tgen, TOPO, dut, GROUP_RANGE_LIST_2, oif, rp_address, SOURCE
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    group_address_list = GROUP_ADDRESS_LIST_1 + GROUP_ADDRESS_LIST_2
    step("r0: Send MLD join for 10 groups")
    result = app_helper.run_join("r0", group_address_list, "r1")
    assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    step("r1: Verify MLD groups")
    dut = "r1"
    intf_r1_r0 = TOPO["routers"]["r1"]["links"]["r0"]["interface"]
    result = verify_mld_groups(tgen, dut, intf_r1_r0, group_address_list)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r5: Send multicast traffic for group {}".format(group_address_list))
    SOURCE_ADDRESS = TOPO["routers"]["r5"]["links"]["r3"]["ipv6"].split("/")[0]
    result = app_helper.run_traffic("r5", group_address_list, "r3")
    assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    step("r1: Verify (*, G) upstream IIF interface")
    dut = "r1"
    iif1 = TOPO["routers"]["r1"]["links"]["r2"]["interface"]
    iif2 = TOPO["routers"]["r1"]["links"]["r4"]["interface"]

    for _iif, _group in zip([iif1, iif2], [GROUP_ADDRESS_LIST_1, GROUP_ADDRESS_LIST_2]):
        result = verify_upstream_iif(tgen, dut, _iif, STAR, _group)
        assert result is True, ASSERT_MSG.format(tc_name, result)

        step("r1: Verify (*, G) upstream join state and join timer")
        result = verify_join_state_and_timer(
            tgen, dut, _iif, STAR, _group, addr_type="ipv6"
        )
        assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (*, G) ip mroutes")
    iif = TOPO["routers"]["r1"]["links"]["r2"]["interface"]
    oif = TOPO["routers"]["r1"]["links"]["r0"]["interface"]
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS_LIST_1, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (S, G) upstream IIF interface")
    iif = TOPO["routers"]["r1"]["links"]["r3"]["interface"]
    result = verify_upstream_iif(tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_1)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_1, addr_type="ipv6"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (S, G) ip mroutes")
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_1, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (*, G) upstream IIF interface")
    dut = "r2"
    iif = "lo"
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS_LIST_1)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (*, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, STAR, GROUP_ADDRESS_LIST_1, addr_type="ipv6"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (*, G) ip mroutes")
    oif = TOPO["routers"]["r2"]["links"]["r1"]["interface"]
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS_LIST_1, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (S, G) upstream IIF interface")
    iif = TOPO["routers"]["r2"]["links"]["r3"]["interface"]
    result = verify_upstream_iif(
        tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_1, joinState="NotJoined"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen,
        dut,
        iif,
        SOURCE_ADDRESS,
        GROUP_ADDRESS_LIST_1,
        addr_type="ipv6",
        expected=False,
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "r2: (S,G) upstream state is joined and join timer is running\n Error: {}".format(
            tc_name, result
        )
    )

    step("r2: Verify (S, G) ip mroutes")
    oif = "none"
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_1, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r3: Verify (S, G) upstream IIF interface")
    dut = "r3"
    iif = TOPO["routers"]["r3"]["links"]["r5"]["interface"]
    result = verify_upstream_iif(tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_1)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r3: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen,
        dut,
        iif,
        SOURCE_ADDRESS,
        GROUP_ADDRESS_LIST_1,
        addr_type="ipv6",
        expected=False,
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "r3: (S,G) upstream state is joined and join timer is running\n Error: {}".format(
            tc_name, result
        )
    )

    step("r3: Verify (S, G) ip mroutes")
    oif = TOPO["routers"]["r3"]["links"]["r1"]["interface"]
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_1, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (*, G) upstream IIF interface")
    dut = "r1"
    iif = TOPO["routers"]["r1"]["links"]["r4"]["interface"]
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS_LIST_2)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (*, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, STAR, GROUP_ADDRESS_LIST_2, addr_type="ipv6"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (*, G) ip mroutes")
    oif = TOPO["routers"]["r1"]["links"]["r0"]["interface"]
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS_LIST_2, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (S, G) upstream IIF interface")
    iif = TOPO["routers"]["r1"]["links"]["r3"]["interface"]
    result = verify_upstream_iif(tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_2)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_2, addr_type="ipv6"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (S, G) ip mroutes")
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_2, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r4: Verify (*, G) upstream IIF interface")
    dut = "r4"
    iif = "lo"
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS_LIST_2)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r4: Verify (*, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, STAR, GROUP_ADDRESS_LIST_2, addr_type="ipv6"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r4: Verify (*, G) ip mroutes")
    oif = TOPO["routers"]["r4"]["links"]["r1"]["interface"]
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS_LIST_2, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r4: Verify (S, G) upstream IIF interface")
    iif = TOPO["routers"]["r4"]["links"]["r3"]["interface"]
    result = verify_upstream_iif(
        tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_2, joinState="NotJoined"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r4: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen,
        dut,
        iif,
        SOURCE_ADDRESS,
        GROUP_ADDRESS_LIST_2,
        addr_type="ipv6",
        expected=False,
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "r4: (S,G) upstream state is joined and join timer is running\n Error: {}".format(
            tc_name, result
        )
    )

    step("r4: Verify (S, G) ip mroutes")
    oif = "none"
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_2, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r3: Verify (S, G) upstream IIF interface")
    dut = "r3"
    iif = TOPO["routers"]["r3"]["links"]["r5"]["interface"]
    result = verify_upstream_iif(tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_2)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r3: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen,
        dut,
        iif,
        SOURCE_ADDRESS,
        GROUP_ADDRESS_LIST_2,
        addr_type="ipv6",
        expected=False,
    )
    assert result is not True, "Testcase {} :Failed \n Error: {}".format(
        tc_name, result
    )

    step("r3: Verify (S, G) ip mroutes")
    oif = TOPO["routers"]["r3"]["links"]["r1"]["interface"]
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_2, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("Delete RP configuration")
    input_dict = {
        "r2": {
            "pim6": {
                "rp": [
                    {
                        "rp_addr": TOPO["routers"]["r2"]["links"]["lo"]["ipv6"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_LIST_1,
                        "delete": True,
                    }
                ]
            }
        },
        "r4": {
            "pim6": {
                "rp": [
                    {
                        "rp_addr": TOPO["routers"]["r4"]["links"]["lo"]["ipv6"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_LIST_2,
                        "delete": True,
                    }
                ]
            }
        },
    }

    result = create_pim_config(tgen, TOPO, input_dict)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("r1, r2, r3, r4: Re-configure RP")
    input_dict = {
        "r2": {
            "pim6": {
                "rp": [
                    {
                        "rp_addr": TOPO["routers"]["r2"]["links"]["lo"]["ipv6"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_LIST_1,
                    }
                ]
            }
        },
        "r4": {
            "pim6": {
                "rp": [
                    {
                        "rp_addr": TOPO["routers"]["r4"]["links"]["lo"]["ipv6"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_LIST_2,
                    }
                ]
            }
        },
    }

    result = create_pim_config(tgen, TOPO, input_dict)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("r1: Shut/No Shut the interfacesfrom R1 to R2, R4 and R0")
    dut = "r1"
    intf1 = TOPO["routers"]["r1"]["links"]["r2"]["interface"]
    intf2 = TOPO["routers"]["r1"]["links"]["r4"]["interface"]
    intf3 = TOPO["routers"]["r1"]["links"]["r0"]["interface"]
    for intf in [intf1, intf2, intf3]:
        shutdown_bringup_interface(tgen, dut, intf, False)
        shutdown_bringup_interface(tgen, dut, intf, True)

    step("r1: Verify (*, G) upstream IIF interface")
    dut = "r1"
    iif = TOPO["routers"]["r1"]["links"]["r2"]["interface"]
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS_LIST_1)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (*, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, STAR, GROUP_ADDRESS_LIST_1, addr_type="ipv6"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (*, G) ip mroutes")
    oif = TOPO["routers"]["r1"]["links"]["r0"]["interface"]
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS_LIST_1, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (S, G) upstream IIF interface")
    iif = TOPO["routers"]["r1"]["links"]["r3"]["interface"]
    result = verify_upstream_iif(tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_1)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_1, addr_type="ipv6"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (S, G) ip mroutes")
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_1, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (*, G) upstream IIF interface")
    dut = "r2"
    iif = "lo"
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS_LIST_1)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (*, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, STAR, GROUP_ADDRESS_LIST_1, addr_type="ipv6"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (*, G) ip mroutes")
    oif = TOPO["routers"]["r2"]["links"]["r1"]["interface"]
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS_LIST_1, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (S, G) upstream IIF interface")
    iif = TOPO["routers"]["r2"]["links"]["r3"]["interface"]
    result = verify_upstream_iif(
        tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_1, joinState="NotJoined"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen,
        dut,
        iif,
        SOURCE_ADDRESS,
        GROUP_ADDRESS_LIST_1,
        addr_type="ipv6",
        expected=False,
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "r2: (S,G) upstream state is joined and join timer is running\n Error: {}".format(
            tc_name, result
        )
    )

    step("r2: Verify (S, G) ip mroutes")
    oif = "none"
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_1, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r3: Verify (S, G) upstream IIF interface")
    dut = "r3"
    iif = TOPO["routers"]["r3"]["links"]["r5"]["interface"]
    result = verify_upstream_iif(tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_1)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r3: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen,
        dut,
        iif,
        SOURCE_ADDRESS,
        GROUP_ADDRESS_LIST_1,
        addr_type="ipv6",
        expected=False,
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "r3: (S,G) upstream state is joined and join timer is running\n Error: {}".format(
            tc_name, result
        )
    )

    step("r3: Verify (S, G) ip mroutes")
    oif = TOPO["routers"]["r3"]["links"]["r1"]["interface"]
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_1, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (*, G) upstream IIF interface")
    dut = "r1"
    iif = TOPO["routers"]["r1"]["links"]["r4"]["interface"]
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS_LIST_2)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (*, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, STAR, GROUP_ADDRESS_LIST_2, addr_type="ipv6"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (*, G) ip mroutes")
    oif = TOPO["routers"]["r1"]["links"]["r0"]["interface"]
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS_LIST_2, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (S, G) upstream IIF interface")
    iif = TOPO["routers"]["r1"]["links"]["r3"]["interface"]
    result = verify_upstream_iif(tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_2)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_2, addr_type="ipv6"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (S, G) ip mroutes")
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_2, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r4: Verify (*, G) upstream IIF interface")
    dut = "r4"
    iif = "lo"
    result = verify_upstream_iif(tgen, dut, iif, STAR, GROUP_ADDRESS_LIST_2)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r4: Verify (*, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen, dut, iif, STAR, GROUP_ADDRESS_LIST_2, addr_type="ipv6"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r4: Verify (*, G) ip mroutes")
    oif = TOPO["routers"]["r4"]["links"]["r1"]["interface"]
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS_LIST_2, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r4: Verify (S, G) upstream IIF interface")
    iif = TOPO["routers"]["r4"]["links"]["r3"]["interface"]
    result = verify_upstream_iif(
        tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_2, joinState="NotJoined"
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r4: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen,
        dut,
        iif,
        SOURCE_ADDRESS,
        GROUP_ADDRESS_LIST_2,
        addr_type="ipv6",
        expected=False,
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "r4: (S,G) upstream state is joined and join timer is running\n Error: {}".format(
            tc_name, result
        )
    )

    step("r4: Verify (S, G) ip mroutes")
    oif = "none"
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_2, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r3: Verify (S, G) upstream IIF interface")
    dut = "r3"
    iif = TOPO["routers"]["r3"]["links"]["r5"]["interface"]
    result = verify_upstream_iif(tgen, dut, iif, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_2)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r3: Verify (S, G) upstream join state and join timer")
    result = verify_join_state_and_timer(
        tgen,
        dut,
        iif,
        SOURCE_ADDRESS,
        GROUP_ADDRESS_LIST_2,
        addr_type="ipv6",
        expected=False,
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "r3: (S,G) upstream state is joined and join timer is running\n Error: {}".format(
            tc_name, result
        )
    )

    step("r3: Verify (S, G) ip mroutes")
    oif = TOPO["routers"]["r3"]["links"]["r1"]["interface"]
    result = verify_mroutes(tgen, dut, SOURCE_ADDRESS, GROUP_ADDRESS_LIST_2, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    write_test_footer(tc_name)


def test_pim6_delete_RP_shut_noshut_upstream_interface_p1(request):
    """
    Verify RP info and (*,G) mroute after deleting the RP and shut /
    no shut the RPF interface.

    Topology used:
                ________r2_____
                |             |
      iperf     |             |
        r0-----r1-------------r3
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("Creating configuration from JSON")
    reset_config_on_routers(tgen)

    app_helper.stop_all_hosts()

    step("Enable MLD on r1 interface")
    step("Enable the PIM6 on all the interfaces of r1, r2, r3 and r4 routers")
    step("r2: Configure r2 as RP")
    input_dict = {
        "r2": {
            "pim6": {
                "rp": [
                    {
                        "rp_addr": TOPO["routers"]["r2"]["links"]["lo"]["ipv6"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_1,
                    }
                ]
            }
        }
    }

    result = create_pim_config(tgen, TOPO, input_dict)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("r2: verify rp-info")
    dut = "r2"
    oif = "lo"
    rp_address = TOPO["routers"]["r2"]["links"]["lo"]["ipv6"].split("/")[0]
    result = verify_pim_rp_info(
        tgen, TOPO, dut, GROUP_ADDRESS_1, oif, rp_address, SOURCE
    )
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r0: Send MLD join")
    result = app_helper.run_join("r0", GROUP_ADDRESS_1, "r1")
    assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    step("r1: Verify MLD groups")
    dut = "r1"
    intf_r1_r0 = TOPO["routers"]["r1"]["links"]["r0"]["interface"]
    result = verify_mld_groups(tgen, dut, intf_r1_r0, GROUP_ADDRESS_1)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Verify (*, G) ip mroutes created")
    iif = TOPO["routers"]["r1"]["links"]["r2"]["interface"]
    oif = TOPO["routers"]["r1"]["links"]["r0"]["interface"]
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS_1, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r2: Verify (*, G) ip mroutes created")
    dut = "r2"
    iif = "lo"
    oif = TOPO["routers"]["r2"]["links"]["r1"]["interface"]
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS_1, iif, oif)
    assert result is True, ASSERT_MSG.format(tc_name, result)

    step("r1: Delete RP configuration")
    input_dict = {
        "r1": {
            "pim6": {
                "rp": [
                    {
                        "rp_addr": TOPO["routers"]["r2"]["links"]["lo"]["ipv6"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_1,
                        "delete": True,
                    }
                ]
            }
        }
    }

    result = create_pim_config(tgen, TOPO, input_dict)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("r1: Shut/No Shut the interface r1-r2-eth1/r1-r0-eth0 from R1 to R2")
    dut = "r1"
    intf1 = TOPO["routers"]["r1"]["links"]["r2"]["interface"]
    intf2 = TOPO["routers"]["r1"]["links"]["r0"]["interface"]
    for intf in [intf1, intf2]:
        shutdown_bringup_interface(tgen, dut, intf, False)
        shutdown_bringup_interface(tgen, dut, intf, True)

    step("r2: Shut the RP interface lo")
    dut = "r2"
    intf = "lo"
    shutdown_bringup_interface(tgen, dut, intf, False)

    step("r1: Shut the interface r1-r2-eth1/r1-r3-eth2 towards RP")
    intf3 = TOPO["routers"]["r1"]["links"]["r3"]["interface"]
    for intf in [intf1, intf3]:
        shutdown_bringup_interface(tgen, dut, intf, False)

    step("r1: Verify (*, G) ip mroutes cleared")
    dut = "r1"
    iif = TOPO["routers"]["r1"]["links"]["r2"]["interface"]
    oif = TOPO["routers"]["r1"]["links"]["r0"]["interface"]
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS_1, iif, oif, expected=False)
    assert result is not True, (
        "Testcase {} : Failed \n "
        "r1: (*,G) mroutes are not cleared after shut of R1 to R0 link\n Error: {}".format(
            tc_name, result
        )
    )

    step("r2: Verify (*, G) ip mroutes cleared")
    dut = "r2"
    iif = "lo"
    oif = TOPO["routers"]["r2"]["links"]["r1"]["interface"]
    result = verify_mroutes(tgen, dut, STAR, GROUP_ADDRESS_1, iif, oif, expected=False)
    assert result is not True, (
        "Testcase {} : Failed \n "
        "r2: (*,G) mroutes are not cleared after shut of R1 to R0 link\n Error: {}".format(
            tc_name, result
        )
    )

    write_test_footer(tc_name)


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



# SPDX-License-Identifier: NONE
# 2020 by David Lamparter, placed in the public domain.

import sys
import os
import re
import argparse

wrap_res = [
    (re.compile(r'(?<!\\n)"\s*\n\s*"', re.M), r""),
]
pri_res = [
    (re.compile(r'(PRI[udx][0-9]+)\s*\n\s*"', re.M), r'\1"'),
    (re.compile(r'"\s*PRI([udx])32\s*"'), r"\1"),
    (re.compile(r'"\s*PRI([udx])32'), r'\1"'),
    (re.compile(r'"\s*PRI([udx])16\s*"'), r"h\1"),
    (re.compile(r'"\s*PRI([udx])16'), r'h\1"'),
    (re.compile(r'"\s*PRI([udx])8\s*"'), r"hh\1"),
    (re.compile(r'"\s*PRI([udx])8'), r'hh\1"'),
]


def main():
    argp = argparse.ArgumentParser(description="C string mangler")
    argp.add_argument("--unwrap", action="store_const", const=True)
    argp.add_argument("--pri8-16-32", action="store_const", const=True)
    argp.add_argument("files", type=str, nargs="+")
    args = argp.parse_args()

    regexes = []
    if args.unwrap:
        regexes.extend(wrap_res)
    if args.pri8_16_32:
        regexes.extend(pri_res)
    if len(regexes) == 0:
        sys.stderr.write("no action selected to execute\n")
        sys.exit(1)

    l = 0

    for fn in args.files:
        sys.stderr.write(fn + "\033[K\r")
        with open(fn, "r") as ifd:
            data = ifd.read()

        newdata = data
        n = 0
        for regex, repl in regexes:
            newdata, m = regex.subn(repl, newdata)
            n += m

        if n > 0:
            sys.stderr.write("changed: %s\n" % fn)
            with open(fn + ".new", "w") as ofd:
                ofd.write(newdata)
            os.rename(fn + ".new", fn)
            l += 1

    sys.stderr.write("%d files changed.\n" % (l))


main()



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2022 by VMware, Inc. ("VMware")
# Used Copyright (c) 2018 by Network Device Education Foundation,
# Inc. ("NetDEF") in this file.
#

import sys
import time
import pytest
import inspect
import os

"""Following tests are covered to test bgp admin distance functionality.
TC_5:
        Verify bgp admin distance functionality when static route is configured
        same as bgp learnt route in user vrf.

TC_6:   Verify bgp admin distance functionality with ECMP in user vrf.

TC_7:
        Verify bgp admin distance functionality when routes are
        imported between VRFs.
"""

#################################
# TOPOLOGY
#################################
"""

                    +-------+
         +--------- |  R2   |
         |          +-------+
         |iBGP           |
     +-------+           |
     |  R1   |           |iBGP
     +-------+           |
         |               |
         |    iBGP   +-------+   eBGP   +-------+
         +---------- |  R3   |----------|  R4   |
                     +-------+          +-------+
                        |
                        |eBGP
                        |
                    +-------+
                    |  R5   |
                    +-------+


"""

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))
sys.path.append(os.path.join(CWD, "../lib/"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import Topogen, get_topogen

# Required to instantiate the topology builder class.
from lib.common_config import (
    start_topology,
    write_test_header,
    step,
    write_test_footer,
    create_static_routes,
    verify_rib,
    check_address_types,
    reset_config_on_routers,
    check_router_status,
)
from lib.topolog import logger
from lib.bgp import (
    verify_bgp_convergence,
    create_router_bgp,
    verify_best_path_as_per_admin_distance,
)

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import Topogen, get_topogen
from lib.topojson import build_config_from_json
from lib.topolog import logger

# Global variables
topo = None
bgp_convergence = False
pytestmark = [pytest.mark.bgpd, pytest.mark.staticd]

NETWORK = {
    "ipv4": [
        "192.168.20.1/32",
        "192.168.20.2/32",
        "192.168.21.1/32",
        "192.168.21.2/32",
        "192.168.22.1/32",
        "192.168.22.2/32",
    ],
    "ipv6": [
        "fc07:50::1/128",
        "fc07:50::2/128",
        "fc07:150::1/128",
        "fc07:150::2/128",
        "fc07:1::1/128",
        "fc07:1::2/128",
    ],
}
ADDR_TYPES = check_address_types()


def setup_module(mod):
    """
    Sets up the pytest environment

    * `mod`: module name
    """

    global topo

    testsuite_run_time = time.asctime(time.localtime(time.time()))
    logger.info("Testsuite start time: {}".format(testsuite_run_time))
    logger.info("=" * 40)

    logger.info("Running setup_module to create topology")

    # This function initiates the topology build with Topogen...
    json_file = "{}/bgp_admin_dist_vrf.json".format(CWD)
    tgen = Topogen(json_file, mod.__name__)
    global topo
    topo = tgen.json_topo

    # Starting topology, create tmp files which are loaded to routers
    #  to start deamons and then start routers
    start_topology(tgen)

    # Creating configuration from JSON
    build_config_from_json(tgen, topo)

    # Checking BGP convergence
    global bgp_convergence
    global ADDR_TYPES

    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Api call verify whether BGP is converged
    bgp_convergence = verify_bgp_convergence(tgen, topo)
    assert bgp_convergence is True, "setup_module :Failed \n Error:" " {}".format(
        bgp_convergence
    )
    logger.info("Running setup_module() done")


def teardown_module(mod):
    """teardown_module.

    Teardown the pytest environment.
    * `mod`: module name
    """
    logger.info("Running teardown_module to delete topology")
    tgen = get_topogen()

    # Stop toplogy and Remove tmp files
    tgen.stop_topology()

    logger.info(
        "Testsuite end time: {}".format(time.asctime(time.localtime(time.time())))
    )
    logger.info("=" * 40)


#####################################################
# Tests starting
#####################################################


def test_bgp_admin_distance_ebgp_vrf_p0():
    """
    TC: 5
    Verify bgp admin distance functionality when static route is
    configured same as ebgp learnt route
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip("skipping test case because of BGP Convergence failure at setup")

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    if tgen.routers_have_failure():
        check_router_status(tgen)

    step("Configure base config as per the topology")
    reset_config_on_routers(tgen)

    step("Configure bgp admin distance 200 with CLI in dut.")

    input_dict_1 = {
        "r3": {
            "bgp": [
                {
                    "vrf": "RED",
                    "local_as": 100,
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "distance": {"ebgp": 200, "ibgp": 200, "local": 200}
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "distance": {"ebgp": 200, "ibgp": 200, "local": 200}
                            }
                        },
                    },
                }
            ]
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify bgp routes have admin distance of 200 in dut.")
    # Verifying best path
    dut = "r3"
    attribute = "admin_distance"

    input_dict = {
        "ipv4": {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK["ipv4"][0],
                        "admin_distance": 200,
                        "vrf": "RED",
                    },
                    {
                        "network": NETWORK["ipv4"][1],
                        "admin_distance": 200,
                        "vrf": "RED",
                    },
                ]
            }
        },
        "ipv6": {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK["ipv6"][0],
                        "admin_distance": 200,
                        "vrf": "RED",
                    },
                    {
                        "network": NETWORK["ipv6"][1],
                        "admin_distance": 200,
                        "vrf": "RED",
                    },
                ]
            }
        },
    }

    for addr_type in ADDR_TYPES:
        result = verify_best_path_as_per_admin_distance(
            tgen, addr_type, dut, input_dict[addr_type], attribute, vrf="RED"
        )
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Modify the admin distance value to 150.")

    input_dict_1 = {
        "r3": {
            "bgp": [
                {
                    "local_as": 100,
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "distance": {"ebgp": 150, "ibgp": 150, "local": 150}
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "distance": {"ebgp": 150, "ibgp": 150, "local": 150}
                            }
                        },
                    },
                }
            ]
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify bgp routes have admin distance of 150 in dut.")
    # Verifying best path
    dut = "r3"
    attribute = "admin_distance"

    input_dict = {
        "ipv4": {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK["ipv4"][0],
                        "admin_distance": 150,
                        "vrf": "RED",
                    },
                    {
                        "network": NETWORK["ipv4"][1],
                        "admin_distance": 150,
                        "vrf": "RED",
                    },
                ]
            }
        },
        "ipv6": {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK["ipv6"][0],
                        "admin_distance": 150,
                        "vrf": "RED",
                    },
                    {
                        "network": NETWORK["ipv6"][1],
                        "admin_distance": 150,
                        "vrf": "RED",
                    },
                ]
            }
        },
    }

    for addr_type in ADDR_TYPES:
        result = verify_best_path_as_per_admin_distance(
            tgen, addr_type, dut, input_dict[addr_type], attribute, vrf="RED"
        )
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Un configure the admin distance value on DUT")

    input_dict_1 = {
        "r3": {
            "bgp": [
                {
                    "local_as": 100,
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "distance": {
                                    "ebgp": 150,
                                    "ibgp": 150,
                                    "local": 150,
                                    "delete": True,
                                }
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "distance": {
                                    "ebgp": 150,
                                    "ibgp": 150,
                                    "local": 150,
                                    "delete": True,
                                }
                            }
                        },
                    },
                }
            ]
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify bgp routes have default admin distance in dut.")
    # Verifying best path
    dut = "r3"
    attribute = "admin_distance"

    input_dict = {
        "ipv4": {
            "r3": {
                "static_routes": [
                    {"network": NETWORK["ipv4"][0], "admin_distance": 20, "vrf": "RED"},
                    {"network": NETWORK["ipv4"][1], "admin_distance": 20, "vrf": "RED"},
                ]
            }
        },
        "ipv6": {
            "r3": {
                "static_routes": [
                    {"network": NETWORK["ipv6"][0], "admin_distance": 20, "vrf": "RED"},
                    {"network": NETWORK["ipv6"][1], "admin_distance": 20, "vrf": "RED"},
                ]
            }
        },
    }

    for addr_type in ADDR_TYPES:
        result = verify_best_path_as_per_admin_distance(
            tgen, addr_type, dut, input_dict[addr_type], attribute, vrf="RED"
        )
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Configure static route  Without any admin distance")

    for addr_type in ADDR_TYPES:
        # Create Static routes
        input_dict = {
            "r3": {
                "static_routes": [
                    {"network": NETWORK[addr_type], "next_hop": "Null0", "vrf": "RED"}
                ]
            }
        }

        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Verify that zebra selects static route.")
    protocol = "static"
    # dual stack changes
    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {"network": NETWORK[addr_type], "next_hop": "Null0", "vrf": "RED"}
                ]
            }
        }
        result4 = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    step("Configure static route  with admin distance of 253")
    for addr_type in ADDR_TYPES:
        # Create Static routes
        input_dict = {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": "Null0",
                        "admin_distance": 253,
                        "vrf": "RED",
                    }
                ]
            }
        }

        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Verify that zebra selects bgp route.")
    protocol = "bgp"

    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": "Null0",
                        "admin_distance": 253,
                        "vrf": "RED",
                    }
                ]
            }
        }
        result4 = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    step("Configure admin distance of 254 in bgp for route  .")

    input_dict_1 = {
        "r3": {
            "bgp": [
                {
                    "local_as": 100,
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "distance": {"ebgp": 254, "ibgp": 254, "local": 254}
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "distance": {"ebgp": 254, "ibgp": 254, "local": 254}
                            }
                        },
                    },
                }
            ]
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify that zebra selects static route.")
    protocol = "static"
    # dual stack changes
    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": "Null0",
                        "admin_distance": 253,
                        "vrf": "RED",
                    }
                ]
            }
        }

        result4 = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    step("Configure admin distance of 255 in bgp for route  in vrf red")

    input_dict_1 = {
        "r3": {
            "bgp": [
                {
                    "local_as": 100,
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "distance": {"ebgp": 255, "ibgp": 255, "local": 255}
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "distance": {"ebgp": 255, "ibgp": 255, "local": 255}
                            }
                        },
                    },
                }
            ]
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify that zebra selects static route.")
    protocol = "static"
    # dual stack changes
    for addr_type in ADDR_TYPES:
        input_dict = {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": "Null0",
                        "admin_distance": 253,
                        "vrf": "RED",
                    }
                ]
            }
        }

        result4 = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    step("Delete the static route.")
    for addr_type in ADDR_TYPES:
        # Create Static routes
        input_dict = {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": "Null0",
                        "admin_distance": 253,
                        "delete": True,
                        "vrf": "RED",
                    }
                ]
            }
        }

        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Verify that zebra selects bgp route.")
    protocol = "bgp"
    # dual stack changes
    for addr_type in ADDR_TYPES:
        result4 = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    write_test_footer(tc_name)


def test_bgp_admin_distance_ebgp_with_imported_rtes_vrf_p0():
    """
    TC: 5
    Verify bgp admin distance functionality when static route is configured
    same as bgp learnt route in user vrf.
    """
    tgen = get_topogen()
    global bgp_convergence

    if bgp_convergence is not True:
        pytest.skip("skipping test case because of BGP Convergence failure at setup")

    # test case name
    tc_name = inspect.stack()[0][3]
    write_test_header(tc_name)
    if tgen.routers_have_failure():
        check_router_status(tgen)

    step("Configure base config as per the topology")
    reset_config_on_routers(tgen)
    step("Configure bgp admin distance 200 with CLI in dut.")
    step(" Import route from vrf to default vrf")
    input_dict_1 = {
        "r3": {
            "bgp": [
                {
                    "vrf": "RED",
                    "local_as": 100,
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "distance": {"ebgp": 200, "ibgp": 200, "local": 200}
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "distance": {"ebgp": 200, "ibgp": 200, "local": 200}
                            }
                        },
                    },
                },
                {
                    "local_as": 100,
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "distance": {"ebgp": 200, "ibgp": 200, "local": 200},
                                "import": {"vrf": "RED"},
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "distance": {"ebgp": 200, "ibgp": 200, "local": 200},
                                "import": {
                                    "vrf": "RED",
                                },
                            }
                        },
                    },
                },
            ]
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify bgp routes have admin distance of 200 in dut.")
    # Verifying best path
    dut = "r3"
    attribute = "admin_distance"

    input_dict = {
        "ipv4": {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK["ipv4"][0],
                        "admin_distance": 200,
                        "vrf": "RED",
                    },
                    {
                        "network": NETWORK["ipv4"][1],
                        "admin_distance": 200,
                        "vrf": "RED",
                    },
                ]
            }
        },
        "ipv6": {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK["ipv6"][0],
                        "admin_distance": 200,
                        "vrf": "RED",
                    },
                    {
                        "network": NETWORK["ipv6"][1],
                        "admin_distance": 200,
                        "vrf": "RED",
                    },
                ]
            }
        },
    }

    for addr_type in ADDR_TYPES:
        result = verify_best_path_as_per_admin_distance(
            tgen, addr_type, dut, input_dict[addr_type], attribute, vrf="RED"
        )
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step(
        "Verify that routes are getting imported without any issues and "
        "routes are calculated and installed in rib."
    )

    input_dict = {
        "ipv4": {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK["ipv4"][0],
                        "admin_distance": 200,
                    },
                    {
                        "network": NETWORK["ipv4"][1],
                        "admin_distance": 200,
                    },
                ]
            }
        },
        "ipv6": {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK["ipv6"][0],
                        "admin_distance": 200,
                    },
                    {
                        "network": NETWORK["ipv6"][1],
                        "admin_distance": 200,
                    },
                ]
            }
        },
    }

    step("Verify that zebra selects bgp route.")
    protocol = "bgp"
    # dual stack changes
    for addr_type in ADDR_TYPES:
        result4 = verify_rib(tgen, addr_type, dut, input_dict, protocol=protocol)
        assert result4 is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result4
        )

    step(" Un configure import route vrf red inside default vrf.")
    input_dict_1 = {
        "r3": {
            "bgp": [
                {
                    "vrf": "RED",
                    "local_as": 100,
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "distance": {"ebgp": 200, "ibgp": 200, "local": 200}
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "distance": {"ebgp": 200, "ibgp": 200, "local": 200}
                            }
                        },
                    },
                },
                {
                    "local_as": 100,
                    "address_family": {
                        "ipv4": {
                            "unicast": {
                                "distance": {"ebgp": 200, "ibgp": 200, "local": 200},
                                "import": {"vrf": "RED", "delete": True},
                            }
                        },
                        "ipv6": {
                            "unicast": {
                                "distance": {"ebgp": 200, "ibgp": 200, "local": 200},
                                "import": {"vrf": "RED", "delete": True},
                            }
                        },
                    },
                },
            ]
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    input_dict = {
        "ipv4": {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK["ipv4"][0],
                        "admin_distance": 200,
                    },
                    {
                        "network": NETWORK["ipv4"][1],
                        "admin_distance": 200,
                    },
                ]
            }
        },
        "ipv6": {
            "r3": {
                "static_routes": [
                    {
                        "network": NETWORK["ipv6"][0],
                        "admin_distance": 200,
                    },
                    {
                        "network": NETWORK["ipv6"][1],
                        "admin_distance": 200,
                    },
                ]
            }
        },
    }

    step("Verify that route withdrawal happens properly.")
    protocol = "bgp"
    # dual stack changes
    for addr_type in ADDR_TYPES:
        result4 = verify_rib(
            tgen,
            addr_type,
            dut,
            input_dict[addr_type],
            protocol=protocol,
            expected=False,
        )
        assert (
            result4 is not True
        ), "Testcase {} : Failed \n Route is not withdrawn. Error: {}".format(
            tc_name, result4
        )

    write_test_footer(tc_name)


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))
